#include "OpenGLtexture.h"

TextureClass texture[128];
int CurTexture = 0;
int MaxTextures = 127;
int Filter;

/*************************************************************************
SetTextureFilter(int newfilter)

  This function is for setting the texture filter on all textures in
  memory. It's good for when you add options to your game and want to
  give the user the ability to turn on/off filtering.
  
	This function also runs the ReSetupTextures() function, which loops
	through all textures and sets them up according to the current Filter.
	
*************************************************************************/
void SetTextureFilter(int newfilter) 
{
	if (newfilter >= 0 && newfilter <= 2) 
	{
		Filter = newfilter;
		ReSetupTextures();
	}
}

/*************************************************************************
int LoadTexture(char *filename, bool ColorIsAvg, bool AlphaIsAvg)

  Loads a texture and returns the OpenGL index reference to that
  texture. ColorIsAvg and AlphaIsAvg tell our LoadBitmapTexture()
  function how to load the texture.
  
*************************************************************************/
int LoadTexture(const char *filename, bool ColorIsAvg, bool AlphaIsAvg) 
{
	int a;
	
	a = CurTexture;
	
	memset(&texture[a], 0, sizeof(texture[a]));
	LoadBitmapTexture(&texture[a], filename, ColorIsAvg, AlphaIsAvg, false);
	texture[a].Id = a;
	SetupTexture(&texture[a]);
	
	CurTexture++;
	return CurTexture - 1;
}

/*************************************************************************
void ReSetupTextures(), void ReSetupTexture()

  Loops through textures that have been loaded and sets them up again.
  This is for when you call change Bilinear, MipMap, or Repeat and you
  want that to affect all textures in memory.
  
	You can do this for a single texture by calling ReSetupTexture().
	
*************************************************************************/
void ReSetupTextures() 
{
	int a=0;
	
	for (a=0; a<CurTexture; a++) 
	{
		SetupTexture(&texture[a]);
	}
}

void ReSetupTexture(int t) 
{
	if (t > CurTexture || t < 0) return;
	SetupTexture(&texture[t]);
}

/*************************************************************************
void LoadBitmapTexture(TextureClass *texture, char *filename,
bool ColorIsAvg, bool AlphaIsAvg, bool Reversed)

  Loads "filename" to the TextureClass "texture".
  
	ColorIsAvg - loads the texture in black and white by using the r g b
	values as an average for each pixel.
	
	  AlphaIsAvg - sets the alpha values for the whole image to the average
	  of the r g b values for each pixel. Used when we load font textures so
	  that the fonts are antialiased.
	  
		Reversed - loads the image from end to beginning, this is used for when
		we load fonts. Bitmaps are oddly stored in reverse, with the first pixels
		being at the end of the file.
		
		  Note: we go in reverse for loading fonts because then the pixels are in
		  the right order. Then we'll use a different function to loop through and
		  chop the font texture into separate textures.
		  
*************************************************************************/
void LoadBitmapTexture(TextureClass *texture, const char *filename, bool ColorIsAvg, bool AlphaIsAvg, bool Reversed) 
{
    char    error_message[256];
	FILE * File;
	BITMAPFILEHEADER FileHeader;
	BITMAPINFOHEADER InfoHeader;
	DWORD Type;
	RGBTRIPLE rgb;
	int i, j, o=0;
	
	// open filename for reading
	File = fopen(filename, "rb");
    
    // Couldn't open file
    if(File == NULL )
    {   
         sprintf(error_message, "Couldn't open texture file - %s", filename );
		 MessageBox( NULL, error_message, 
			         "Ex3 by Ronen and Olga", MB_OK | MB_ICONHAND );
	     return;
    }
	
	// read the header and type, then seek to the next position
	fread(&FileHeader, sizeof(FileHeader), 1, File);
	fread(&Type, sizeof(Type), 1, File);
	fseek(File, sizeof(FileHeader), SEEK_SET);
	
	// read the infoheader
	fread(&InfoHeader, sizeof(InfoHeader), 1, File);
	
	
	// set our texture width and height from what is read in the file
	texture->Width = InfoHeader.biWidth;
	texture->Height = InfoHeader.biHeight;
	
	// allocate enough memory to store the bitmap, then fill the memory with zeros
	// this size is width * height * 4
	// 4 means 1 byte for red, green, blue, and alpha
	texture->Image = (byte *) malloc(texture->Width * texture->Height * 4);
	memset(texture->Image, 0, texture->Width * texture->Height * 4);
	
	
	// if we are in reverse, set our current offset to the very end of the file
	if (Reversed) o = texture->Width * texture->Height * 4;
	
	// loop through loading pixel by pixel
	for (j=0; j < InfoHeader.biHeight; j++) 
	{
		
		// if we're in reverse, go backwards one horizontal line of pixels
		if (Reversed) o -= 2 * (texture->Width * 4);
		
		for (i=0; i < InfoHeader.biWidth; i++) 
		{
			
			// load the r g and b values from the file into an RGBTRIPLE structure
			fread(&rgb, sizeof(rgb), 1, File);
			
			// color is not averaged, so get the pixels normally
			if (!ColorIsAvg) {
				texture->Image[o+0] = rgb.rgbtRed;
				texture->Image[o+1] = rgb.rgbtGreen;
				texture->Image[o+2] = rgb.rgbtBlue;
			}
			// color is the average, so set the r g b values to the average of the three
			else 
			{
				texture->Image[o+0] = (rgb.rgbtRed + rgb.rgbtGreen + rgb.rgbtBlue) / 3;
				texture->Image[o+1] = (rgb.rgbtRed + rgb.rgbtGreen + rgb.rgbtBlue) / 3;
				texture->Image[o+2] = (rgb.rgbtRed + rgb.rgbtGreen + rgb.rgbtBlue) / 3;
			}
			
			// Now we'll set the alpha value for this pixel. if AlphaIsAvg, we average
			// the r g and b to get alpha. Otherwise, check if the pixel is bright pink,
			// if so, set the pixel to completely transparent. If its normal, set alpha
			// to 255, meaning completely transparent.
			if (AlphaIsAvg)
				texture->Image[o+3] = ((rgb.rgbtRed + rgb.rgbtGreen + rgb.rgbtBlue) / 3);
			else if (!AlphaIsAvg && rgb.rgbtRed == 255 && rgb.rgbtGreen == 0 && rgb.rgbtBlue == 255)
				texture->Image[o+3] = 0;
			else
				texture->Image[o+3] = 255;
			
			o += 4;
		}
	}
	
	fclose(File);
}

/*************************************************************************
void SetupTexture(TextureClass *texture)

  Sets the OpenGL texture properties for the supplied texture object.
  This uses the Repeat, Bilinear, and MipMap bool values to tell OpenGL
  how to set up our texture properties.
  
*************************************************************************/
void SetupTexture(TextureClass *texture) 
{
	
	// tell OpenGL we want to use this texture's Id
	glBindTexture(GL_TEXTURE_2D, texture->Id);
	
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	
	if (Filter == TF_NONE) 
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
	else if (Filter == TF_BILINEAR) 
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
	}
	else if (Filter == TF_TRILINEAR) 
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	}
	
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
	glTexImage2D(GL_TEXTURE_2D, 0, 4, texture->Width, texture->Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture->Image );
	if (Filter != TF_NONE) gluBuild2DMipmaps( GL_TEXTURE_2D, 4, texture->Width, texture->Height, GL_RGBA, GL_UNSIGNED_BYTE, texture->Image );
}


/*************************************************************************
int SaveBitmapTexture(const char *filename,BMPStruct &bmpStruct)

  Saves a bitmap from a given BMPStruct into a given filename.
  As so frees the bitmap handle used to create the Bitmap.
  
*************************************************************************/
int SaveBitmapTexture(const char *filename,BMPStruct &bmpStruct)
{
    FILE             *fp;          /* Open file pointer */
    int              size,         /* Size of file */
					 infosize,     /* Size of bitmap info */
				     bitsize;      /* Size of bitmap pixels */
    BITMAPFILEHEADER header;       /* File header */
	
	
    /* Try opening the file; use "wb" mode to write this *binary* file. */
    if ((fp = fopen(filename, "wb")) == NULL)
        return (-1);
	
    /* Figure out the bitmap size */
    if (bmpStruct.info.bmiHeader.biSizeImage == 0)
		bitsize = (bmpStruct.info.bmiHeader.biWidth *
		bmpStruct.info.bmiHeader.biBitCount + 7) / 8 *
		abs(bmpStruct.info.bmiHeader.biHeight);
    else
		bitsize = bmpStruct.info.bmiHeader.biSizeImage;
	
    /* Figure out the header size */
    infosize = sizeof(BITMAPINFOHEADER);
    switch (bmpStruct.info.bmiHeader.biCompression)
	{
	case BI_BITFIELDS :
		infosize += 12; /* Add 3 RGB doubleword masks */
		if (bmpStruct.info.bmiHeader.biClrUsed == 0)
			break;
	case BI_RGB :
		if (bmpStruct.info.bmiHeader.biBitCount > 8 &&
			bmpStruct.info.bmiHeader.biClrUsed == 0)
			break;
	case BI_RLE8 :
	case BI_RLE4 :
		if (bmpStruct.info.bmiHeader.biClrUsed == 0)
			infosize += (1 << bmpStruct.info.bmiHeader.biBitCount) * 4;
		else
			infosize += bmpStruct.info.bmiHeader.biClrUsed * 4;
		break;
	}
	
    size = sizeof(BITMAPFILEHEADER) + infosize + bitsize;
	
    /* Write the file header, bitmap information, and bitmap pixel data... */
    header.bfType      = 'MB'; /* Non-portable... sigh */
    header.bfSize      = size;
    header.bfReserved1 = 0;
    header.bfReserved2 = 0;
    header.bfOffBits   = sizeof(BITMAPFILEHEADER) + infosize;
	
    if (fwrite(&header, 1, sizeof(BITMAPFILEHEADER), fp) < sizeof(BITMAPFILEHEADER))
	{
        /* Couldn't write the file header - return... */
        fclose(fp);
        return (-1);
	}
	
    if ((int)fwrite(&(bmpStruct.info), 1, infosize, fp) < infosize)
	{
        /* Couldn't write the bitmap header - return... */
        fclose(fp);
        return (-1);
	}
	
    if ((int)fwrite(bmpStruct.bits, 1, bitsize, fp) < bitsize)
	{
        /* Couldn't write the bitmap - return... */
        fclose(fp);
        return (-1);
	}
	
    /* OK, everything went fine - return... */
    fclose(fp);
	
	/* It is a good time to delete the bitmap allocated in make bitmap */
	if (bmpStruct.bitmap != NULL)
		DeleteObject(bmpStruct.bitmap);

    return (0);
}
