// Point3D.h: interface for the Point class.
// 
//////////////////////////////////////////////////////////////////////

#ifndef POINT3D_H
#define POINT3D_H

#include <math.h> // for sqrt, cos, sin
#include "Real.h"
#include "Serializable.h"

class Point3D;
/** A convinient definition that makes vectors compatible with points.
*/
typedef Point3D Vector3D;

/** this class represents points and vectors in a three-dimensional space.
* The implementation is not hidden, since we rely on it for
* reasons of efficiency.
*/
class Point3D //: public Serializable
{
public:
	/// Constructs and initializes a point to (0,0,0).
	Point3D();
//	virtual ~Point3D() {}

	/** Constructs and initializes a point to the specified values.
	*/
	Point3D(real x, real y, real z);
	
	/// Returns the X coordinate of the point.
	real GetX() const { return m_coordinates[X]; }
	/// Returns the Y coordinate of the point.
	real GetY() const { return m_coordinates[Y]; }
	/// Returns the Z coordinate of the point.
	real GetZ() const { return m_coordinates[Z]; }
	/// Returns the ith coordinate of the point
	real GetCo(int i) { return m_coordinates[i]; }

	const real *GetCoordArray() const { return m_coordinates; }
	
	/// Sets the X coordinate of the point.to the specified value.
	void SetX(real x) { m_coordinates[X] = x; }
	/// Sets the X coordinate of the point.to the specified value.
	void SetY(real y) { m_coordinates[Y] = y; }
	/// Sets the X coordinate of the point.to the specified value.
	void SetZ(real z) { m_coordinates[Z] = z; }
	/// Sets the ith coordinate of the point.to the specified value.
	void SetCo(int i, real val) { m_coordinates[i] = val; }

	/// Returns the euclidian norm of the vector : sqrt(X^2 + Y^2 + Z^2).
	real Size() const { return (real)sqrt(SquareOfSize()); }

	/** Returns the square of the vector's size. 
	*/
	real SquareOfSize() const;
	
	/// Reduces the vector to a unit size vector.
	void Normalize() 
	{ 	
		real size=Size(); 
		if (size>EPSILON) 
		{
			operator*=( real(1) / Size() ); 
		}
	}

	/** Computes the square of the distance between two points
	*/
	static real SquareOfDistance(const Point3D & lhs, const Point3D & rhs);

	/** Computes the distance between two points.
	*/
	static real Distance(const Point3D & lhs, const Point3D & rhs);
	
	/** Computes the scalar product of two vectors.
	* return: the scalar product of the two vectors : x1*x2 + y1*y2.
	* param lhs: a three-dimensional vector.
	* param rhs: a three-dimensional vector.
	*/
	static real ScalarProduct(const Point3D & lhs, const Point3D & rhs)
	{ return lhs.GetX()*rhs.GetX() + lhs.GetY()*rhs.GetY() + lhs.GetZ()*rhs.GetZ(); }
	
	/** Computes the vector product of two vectors.
	* return: the vector product of the two vectors : lhs X rhs.
	* param: lhs a three-dimensional vector.
	* param: rhs a three-dimensional vector.
	*/
	static Vector3D VectorProduct(const Vector3D & lhs, const Vector3D & rhs);

	/** Computes the cosine of the angle between two vectors.
	*/
	static real Angle(const Point3D & lhs, const Point3D & rhs)
	{ return ScalarProduct(lhs, rhs) / (lhs.Size() * rhs.Size()); }

	/** A high performance addition method.
	* param lhs: a vector to add.
	* param rhs: a vector to add.
	* param result: the vector storing the result of the operation.
	*/
	static void FastAdd(Point3D & result, const Point3D & lhs, const Point3D & rhs)
	{
		result.SetX( lhs.GetX() + rhs.GetX() );
		result.SetY( lhs.GetY() + rhs.GetY() );
		result.SetZ( lhs.GetZ() + rhs.GetZ() );
	}
	
	/** A high performance substraction method.
	* param lhs: the vector to substract from.
	* param rhs: the vector to substract.
	* param result: the vector storing the result of the operation.
	*/
	static void FastSub(Point3D & result, const Point3D & lhs, const Point3D & rhs)
	{
		result.SetX( lhs.GetX() - rhs.GetX() );
		result.SetY( lhs.GetY() - rhs.GetY() );
		result.SetZ( lhs.GetZ() - rhs.GetZ() );
	}

	/** Operators */

	/// A self addition operator.
	void operator+=(const Point3D & rhs)
	{ 
		m_coordinates[X] += rhs.GetX();
		m_coordinates[Y] += rhs.GetY();
		m_coordinates[Z] += rhs.GetZ();
	}

	/// A self substraction operator.
	void operator-=(const Point3D & rhs)
	{ 
		m_coordinates[X] -= rhs.GetX();
		m_coordinates[Y] -= rhs.GetY();
		m_coordinates[Z] -= rhs.GetZ();
	}

	/// A self scaling operator.
	void operator*=(real  factor)
	{ 
		m_coordinates[X] *= factor;
		m_coordinates[Y] *= factor;
		m_coordinates[Z] *= factor;
	}


	// For 2D points only!!!
	// Rotation around origin
	void Rotate(real angle);

	/// An equality testing operator
	friend bool operator==(const Point3D & lhs, const Point3D & rhs);
	/// An addition operator
	friend Point3D operator+(const Point3D & lhs, const Point3D & rhs);
	/// A  substraction operator
	friend Point3D operator-(const Point3D & lhs, const Point3D & rhs);
	/// A scaling operator.
	friend Point3D operator*(real  factor       , const Point3D & rhs);
	/// A scaling operator.
	friend Point3D operator*(const Point3D & lhs, real  factor);

	
	/** Serializable interface overrides */

	/// Reads a point from the stream.
	void Read(ifstream & stream);
	/// Writes a point to the stream.
	void Write(ofstream & stream) const;

	/// serializes an object to a stream
	friend ofstream & operator << (ofstream & stream, const Point3D & object)
	{
		object.Write(stream);
		return stream;
	}
	
	/// serializes an object from a stream
	friend ifstream & operator >> (ifstream & stream, Point3D & object)
	{
		object.Read(stream);
		return stream;
	}


private:
	enum {X, Y, Z};

	// The X,Y and Z coordinates
	real m_coordinates[3];
};

// Inline implementations

inline Point3D::Point3D()
{
	m_coordinates[X] = m_coordinates[Y] = m_coordinates[Z] = real(0);
}

inline Point3D::Point3D(real x, real y, real z)
{
	m_coordinates[X] = x;
	m_coordinates[Y] = y;
	m_coordinates[Z] = z;
}

inline real Point3D::SquareOfSize() const
{ 
	return GetX()*GetX() + GetY()*GetY() + GetZ()*GetZ();
}

inline real Point3D::SquareOfDistance(const Point3D & lhs, const Point3D & rhs)
{
	Point3D diff = rhs;
	diff -= lhs;
	return diff.SquareOfSize();
}

inline real Point3D::Distance(const Point3D & lhs, const Point3D & rhs)
{
	Point3D diff = rhs;
	diff -= lhs;
	return diff.Size();
}

inline bool operator==(const Point3D & lhs, const Point3D & rhs)
{
	if (REAL_EQUAL(lhs.GetX(), rhs.GetX()) &&
		REAL_EQUAL(lhs.GetY(), rhs.GetY()) &&
		REAL_EQUAL(lhs.GetZ(), rhs.GetZ()))
		return true;
	return false;
}

inline Point3D operator+(const Point3D & lhs, const Point3D & rhs)
{
	return Point3D(lhs.GetX() + rhs.GetX(),
		lhs.GetY() + rhs.GetY(),
		lhs.GetZ() + rhs.GetZ());
}

inline Point3D operator-(const Point3D & lhs, const Point3D & rhs)
{
	return Point3D(lhs.GetX() - rhs.GetX(),
		lhs.GetY() - rhs.GetY(),
		lhs.GetZ() - rhs.GetZ());
}

inline Point3D operator*(const Point3D & lhs, real  factor)
{
	return (factor*lhs);
}

inline Point3D operator*(real factor, const Point3D & rhs)
{
	return Point3D(factor * rhs.GetX(), 
		factor * rhs.GetY(), 
		factor * rhs.GetZ());
}

inline Vector3D Point3D::VectorProduct(const Vector3D & lhs, const Vector3D & rhs)
{
	real x = lhs.GetY() * rhs.GetZ() - lhs.GetZ() * rhs.GetY();
	real y = lhs.GetZ() * rhs.GetX() - lhs.GetX() * rhs.GetZ();
	real z = lhs.GetX() * rhs.GetY() - lhs.GetY() * rhs.GetX();
	return Point3D(x, y, z);
}

inline void Point3D::Rotate(real angle)
{
	real x = m_coordinates[X]*cos(angle) - m_coordinates[Y]*sin(angle);
	real y = m_coordinates[X]*sin(angle) + m_coordinates[Y]*cos(angle);

	m_coordinates[X] = x;
	m_coordinates[Y] = y;
}

#endif	// POINT3D_H
