/***************************************************/
/*  TexturedMesh.h:                                */
/*  header file for class TexturedMesh.            */
/*  This class describes a triangular mesh         */
/*  with texture .                                 */
/***************************************************/

#ifndef __TEXTUREDMESH_H__
#define __TEXTUREDMESH_H__


#include <fstream.h>
#include <vector>
#include <map>
#include <set>
//#include <hash_set>

#include "Real.h"
#include "Point3D.h"
#include "Point2D.h"
#include "Tripplet.h"
#include "Triangle.h"
#include "Box.h"



#pragma warning( disable : 4786 )
#pragma warning( disable : 4503 )

class TexturedMesh 
{
	
	/***************************************************/
	/*  Type definitions                               */
	/***************************************************/
public:
	
	typedef std::vector<UINT>	IndexVector;
	
protected:
	
	typedef std::pair<UINT, UINT>				IndexPair;
	typedef std::map<IndexPair, IndexVector>	EdgeToIndicesMap;
	typedef std::vector<char>					ColorVector;
	
	
	/***************************************************/
	/*  Static variables                               */
	/***************************************************/
public:
	
	static const UINT	   INVALID_INDEX;
	
	/***************************************************/
	/*  Member variables                               */
	/***************************************************/
protected:
	
	// m_vertices: vector of 3D points (the geometry of the mesh)
	std::vector<Point3D>		m_vertices;
	
	// m_faces: vector of Tripplets (topology). Each Tripplet represents
	// a triangle by containing the indices (in m_vertices) of its 3 vertices.
	std::vector<Tripplet>		m_faces;
	
	// m_vertex_normals: outward normals to vertices of the mesh
	std::vector<Point3D>		m_vertex_normals;
	
	
	// m_face_normals: outward normals to faces of the mesh
	std::vector<Point3D>		m_face_normals;
	
	
	std::vector<Point3D>		m_color_per_vertex;
	std::vector<Point3D>		m_color_per_vertex1;
	
	
	
	// m_uv_coordinates: a vector that holds a set of texture
	// coordinates (u,v pairs). 0 <= u,v <= 1. When mapping
	// texture, the u coordinate is multiplied by texture width
	// and the v coordinate is multiplied by texture height.
	std::vector<Point2D>		m_uv_coordinates;
	
	// m_uv_per_triangle: a vector of length m_faces.size(). For each
	// triangle, the corresponding entry holds Tripplet of indices
	// in m_uv_coordinates vector, which indicate the u,v coordinates
	// associated with the triangle's vertices.
	std::vector<Tripplet>		m_uv_per_triangle;
	
	// m_textures: array of texture filenames.
	// Currently, only 1 texture is supported.
	// Currently, only BMP texture images are supported.
	std::vector<char *>	        m_textures;
	
	
	// m_is2Dmesh: a boolean value. 
	// True if the mesh is flat against the Z axis 
	// (all z coordinates in m_vertices are 0).
	// False otherwise.
	bool						m_is2Dmesh;
	
	
	// The following data members are built by flattening/other
	// algorithms if needed, using the private methods.
	
	// m_edge_list: a table of length m_vertices.size().
	// Entry number N holds vector of vertex indices that
	// are adjacent to vertex N in the mesh. I.e. if N and K
	// are neighbours in the mesh, than K will be present 
	// in the vector of N and N will be present in the vector of K.
	std::vector<IndexVector>	m_edge_list;
	
	
	// m_edge_to_triangles_map: a search tree. The search key
	// is edge, i.e. pair of indices (i,j), such that i<j.
	// The value that corresponds to edge (i,j) is the vector
	// of triangle indices that have this edge. In a legal mesh,
	// there will be at most 2 such triangles for each edge.
	EdgeToIndicesMap			m_edge_to_triangles_map;
	
	
	// m_triangles_adjacency_list: vector of length m_faces.size()
	// Entry number N holds a vector of triangle indices adjacent to 
	// triangle number N. (In a legal mesh, each triangle can have
	// at most 3 neighbours.) In other words, if triangle number N
	// and triangle number K share an edge, then K is in the vector
	// of N and vice versa.
	std::vector<IndexVector>	m_triangles_adjacency_list;
	
	// m_vertex_to_triangles_map: a vector of length m_vertices.size()
	// that for each vertex keeps the triangles it is incident to.
	std::vector<IndexVector>	m_vertex_to_triangles_map;
	
	
	/***************************************************/
	/*   Constructors/Destructors                      */
	/***************************************************/
public:
	
	TexturedMesh() : m_is2Dmesh(false) { };
	
	TexturedMesh(const TexturedMesh & mesh) :
	m_vertices(mesh.m_vertices),
		m_faces(mesh.m_faces),
		m_uv_coordinates(mesh.m_uv_coordinates),
		m_uv_per_triangle(mesh.m_uv_per_triangle),
		m_edge_list(mesh.m_edge_list),
		m_edge_to_triangles_map(mesh.m_edge_to_triangles_map),
		m_triangles_adjacency_list(mesh.m_triangles_adjacency_list),
		m_vertex_to_triangles_map(mesh.m_vertex_to_triangles_map),
		m_is2Dmesh(mesh.m_is2Dmesh),
		m_color_per_vertex(mesh.m_color_per_vertex)
	{
		for (int i = 0; i < mesh.m_textures.size(); ++i)
		{
			char *str = new char[strlen(mesh.m_textures[0]) + 1];
			strcpy(str, mesh.m_textures[0]);
			
			m_textures.push_back(str);
		}
	}
	
	
	~TexturedMesh() { Clear(); };
	
	
	
	/***************************************************/
	/*  Methods				           */
	/***************************************************/
public:
	
	// ReadFromVRML:
	// reads textured mesh from a file
	// returns true on success, false on error. in case of error, the
	// resulting mesh is incomplete and should be deleted.
	// note: call this method only on a newly created TexturedMesh
	// note: not reentrant
	// note: assumes a valid VRML file. result is unpredictable otherwise.
	// note: assumes one mesh, assumes triangular mesh, assumes one texture
	// note by Olga: implemented outside of TexturedMesh.cpp (in VRMLLoader.cpp).
	bool ReadFromVRML(ifstream &fIn);
	
	// WriteToVRML:
	// Writes the mesh to a file, in VRML format.
	bool WriteToVRML(ofstream &out_file) const;
	
	bool WriteToSimpleFormat(ofstream &out_file) const;
	
	// The file format used by the Metro tool
	// geo_source: 0 for m_vertices
	void WriteToSMFFormat(ofstream &out_file) const;
	
	// geo_source: 0 for m_vertices
	bool WriteToPly2Format(ofstream &out_file);
	
	// ReadMeshFromSimpleFormat: 
	// Reads the mesh in the following format:
	// num_vertices  num_triangles
	// vertices (list of 3D points)
	// triangles (list of tripplets)
	void	ReadFromSimpleFormat(ifstream & in_file);
	
	bool    ReadFromPly2Format(ifstream & in_file);
	
	
	UINT GetNumberOfVertices() const;
	UINT GetNumberOfTriangles() const;
	
	real GetTotalEdgeLength();
	
	// AddVertex: adds new vertex. 
	// Returns its index in m_vertices.
	// DO NOT ADD DUPLICATE VERTICES!!!
	UINT AddVertex(const Point3D & new_point);
	
	// AddTripplet: adds new triangle to mesh.
	// Returns its index in m_faces.
	// DO NOT ADD DUPLICATE TRIPPLETS!!!
	UINT AddTripplet(const Tripplet & new_tripplet);
	UINT AddTripplet(UINT ind1, UINT ind2, UINT ind3);
	
	const Point3D  & GetVertex(UINT index) const;
	const Tripplet & GetTripplet(UINT index) const;
	
	// VertexDegree: returns the valency of vertex no.v
	// returns -1 if index v is out of range
	int VertexDegree(UINT v);
	
	Triangle GetTriangle(UINT tri_index) const;
	
	// GetBoundingBox:
	// Returns axes-aligned bounding box of the mesh
	Box GetBoundingBox() const;
	
	// GetCenterOfMass: returns the center point
	// of the mesh (the average of all vertices).
	Point3D GetCenterOfMass() const;
	
	// Translate: translates the mesh by offset_vector
	void Translate(const Point3D & offset_vector);
	
	// Rotate2DMeshAroundOrigin: rotates 2D mesh
	// by angle around the origin (0,0,0).
	// If mesh is 3D, leaves as is.
	void Rotate2DMeshAroundOrigin(real angle);
	
	// Rotate2DMeshToBestBB: 
	// If mesh is 2D, rotates it so its bounding
	// rectangle is the tightest possible.
	// If mesh is not 2D, leaves it as is.
	void Rotate2DMeshToBestBB();
	
	void Scale(real factor);
	
	real GetSurfaceArea() const;
	
	real GetAverageEdgeLength();
	real GetMinEdgeLength();
	
	
	// Creates normals...
	void SetMesh2Render();
	
	// renders the mesh using OpenGL
	void RenderTriangles(bool normal_per_vertex, 
				 int color_coding, 
				 const unsigned char *color = NULL) const;
	
	void RenderPoints() const;
	
	void RenderVertexBalls(float radius, 
						   int render_all, 
						   int num_segments_width,
						   int num_segments_height);
	
	void ReverseAllNormals();
	
	
	std::vector<Point3D>::const_iterator GetConstVertexArray();
	const std::vector<Point3D> & GetVertexArray();

	std::vector<Point3D>::const_iterator GetConstNormalsArray();

	
	// Clear:
	// Removes all contents of the mesh.
	void Clear();
	
	// CleanDegenerateTripplets: removes tripplets
	// from m_faces (and accordingly from m_uv_per_triangle)
	// that have two ore more identical indices.
	// Should be called BEFORE building all edge/triangle maps!!
	void	CleanDegenerateTripplets();
	
	
	// CleanIsolatedVertices: removes vertices that aren't
	// connected to any other vertex. Also updates the m_faces
	// vector accordingly. Destroys all the maps that are
	// affected by this!!!!!
	void	CleanIsolatedVertices();

	int GetNumberOfBoundaryVertices();

	void FixOrientation();
	
	
	
	/***************************************************/
	/*   Auxiliary functions                           */
	/***************************************************/
protected:
	
	// CreateEdgeList: creates the vertex adjacency list m_edge_list.
	void	CreateEdgeList();
	
	// CreateEdgeToTrianglesMap: creates m_edge_to_triangles_map.
	void	CreateEdgeToTrianglesMap();
	
	// CreateEdgeToTrianglesMap: creates m_edge_to_triangles_map.
	void	CreateEdgeToCreaseAngleMap();
	
	// CreateTrianglesAdjacencyList: creates m_triangles_adjacency_list.
	void    CreateTrianglesAdjacencyList();
	
	// CreateVertexToTrianglesMap: creates m_vertex_to_triangles_map.
	void	CreateVertexToTrianglesMap();
	
	// ScanAndAdd: adds index to vector if it's not already in it.
	void	ScanAndAdd(IndexVector & ind_vector, UINT index);
	
	// AddIndexToEdgeMap: auxiliary function. Adds triangle index ind
	// to the vector under key (vertex_ind1,vertex_ind2) in the edge_map.
	void	AddIndexToEdgeMap(EdgeToIndicesMap & edge_map, 
		UINT vertex_ind1, 
		UINT vertex_ind2, 
		UINT ind);
	
	// IsMesh2D: returns true if the mesh is flat
	// against the Z axis (all vertices have z=0).
	// Returns false if mesh isn't flat or if it's empty.
	// Should be called after any alterations of the mesh.
	bool	IsMesh2D() const;
	
	// Checks that vertex i is a boundary vertex
	bool	IsBoundaryVertex(UINT i);
	
	// Checks if i has no neighbours
	bool IsIsolatedVertex(UINT i);
	
	// Checks if i has only one neighbour
	bool IsHangingVertex(UINT i);
	
	// Weird function - returns index of j in the edge list of i
	// don't touch since I don't know where I needed it :-)
	int		Neighbours(UINT i, UINT j);
	
	// returns true if (i,j) is an edge in the mesh graph
	// false otherwise
	bool Are1neighbours(UINT i, UINT j);


	/***********************************************************************************/
	// Adi help functions.
	
	// checks mesh consistency. called by ReadFromVRML().
	// returns true if all checks are ok, false otherwise
	bool assertValid() const;
	
	friend bool vrml_setFaces(TexturedMesh *tm, const char *cszCoordIndex);
	friend bool vrml_setTexFaces(TexturedMesh *tm, const char *cszTexCoordIndex);
	friend bool vrml_setPoints(TexturedMesh *tm, const char *cszPoint);
	friend bool vrml_setUVPoints(TexturedMesh *tm, const char *cszPoint);
	friend bool vrml_setUrl(TexturedMesh *tm, const char *cszUrl);
};


/***************************************************/
/*   Inline implementations                        */
/***************************************************/

// Clear:
// Removes all contents of the mesh.
inline void TexturedMesh::Clear()
{
	m_vertices.clear();
	m_faces.clear();
	m_uv_coordinates.clear();
	m_uv_per_triangle.clear();
	
	for (int i = 0; i < m_textures.size(); ++i)
	{
		delete [] (m_textures[i]);
	}
	m_textures.clear();
	
	m_edge_list.clear();
	m_edge_to_triangles_map.clear();
	m_triangles_adjacency_list.clear();
	m_vertex_to_triangles_map.clear();
	m_vertex_normals.clear();
	m_face_normals.clear();
	m_color_per_vertex.clear();
	m_color_per_vertex1.clear();
}


inline UINT TexturedMesh::GetNumberOfVertices() const
{
	return m_vertices.size();
}

inline UINT TexturedMesh::GetNumberOfTriangles() const
{
	return m_faces.size();
}

// AddVertex: adds new vertex. 
// Returns its index in m_vertices.
// DO NOT ADD DUPLICATE VERTICES!!!
inline UINT TexturedMesh::AddVertex(const Point3D & new_point)
{
	m_vertices.push_back(new_point);
	return (GetNumberOfVertices() - 1);
}

// AddTripplet: adds new triangle to mesh.
// Returns its index in m_faces.
// DO NOT ADD DUPLICATE TRIPPLETS!!!
inline UINT TexturedMesh::AddTripplet(const Tripplet & new_tripplet)
{
	assert(new_tripplet.GetFirstIndex() < GetNumberOfVertices());
	assert(new_tripplet.GetSecondIndex() < GetNumberOfVertices());
	assert(new_tripplet.GetThirdIndex() < GetNumberOfVertices());
	
	m_faces.push_back(new_tripplet);
	return (GetNumberOfTriangles() - 1);
}

inline UINT TexturedMesh::AddTripplet(UINT ind1, UINT ind2, UINT ind3)
{
	assert(ind1 < GetNumberOfVertices());
	assert(ind2 < GetNumberOfVertices());
	assert(ind3 < GetNumberOfVertices());
	return AddTripplet(Tripplet(ind1, ind2, ind3));
}

inline Triangle TexturedMesh::GetTriangle(UINT tri_index) const
{
	assert(tri_index < m_faces.size());
	Tripplet tripplet = m_faces[tri_index];
	return Triangle(m_vertices[tripplet.GetFirstIndex()],
		m_vertices[tripplet.GetSecondIndex()],
		m_vertices[tripplet.GetThirdIndex()]);
	
}

inline const Point3D & TexturedMesh::GetVertex(UINT index) const
{
	assert(index < GetNumberOfVertices());
	return m_vertices[index];
}

inline const Tripplet & TexturedMesh::GetTripplet(UINT index) const
{
	assert(index < GetNumberOfTriangles());
	return m_faces[index];
}

inline std::vector<Point3D>::const_iterator TexturedMesh::GetConstVertexArray()
{
	return m_vertices.begin();
}

inline std::vector<Point3D>::const_iterator TexturedMesh::GetConstNormalsArray()
{
	return m_vertex_normals.begin();
}


inline const std::vector<Point3D> & TexturedMesh::GetVertexArray() {
	return m_vertices;
}


inline int TexturedMesh::GetNumberOfBoundaryVertices() {
	int counter = 0;
	for (UINT i=0; i < m_vertices.size(); ++i) {
		if (IsBoundaryVertex(i))
			counter++;
	}
	
	return counter;
}

#endif // __TEXTUREDMESH_H__
